@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 
-from argparse import ArgumentParser
+from argparse import ArgumentParser, SUPPRESS
 from collections import OrderedDict
 import os
 import re
@@ -8,6 +8,7 @@ import sys
 
 from ._version import __version__
 from .exceptions import MissingGenomeConfigError, MissingFolderError
+from .asset_build_packages import asset_build_packages
 
 import logmuse
 import pypiper
@@ -25,8 +26,11 @@ PULL_CMD = "pull"
 LIST_LOCAL_CMD = "list"
 LIST_REMOTE_CMD = "listr"
 GET_ASSET_CMD = "seek"
+INSERT_CMD = "add"
 
 
+BUILD_SPECIFIC_ARGS = ('fasta', 'gtf', 'context', 'annogene')
+
 # This establishes the API with the server
 refgenie_server_api = {
     "list_available_genomes": "/genomes",
@@ -69,11 +73,12 @@ def build_argparser():
 
     subparser_messages = {
         INIT_CMD: "Initialize a genome configuration.",
-        LIST_LOCAL_CMD: "List available local genomes.",
-        LIST_REMOTE_CMD: "List available genomes and assets on server.",
+        LIST_LOCAL_CMD: "List available local assets.",
+        LIST_REMOTE_CMD: "List available remote assets.",
         PULL_CMD: "Download assets.",
-        BUILD_CMD: "Build genome assets",
-        GET_ASSET_CMD: "Get the path to a local asset"
+        BUILD_CMD: "Build genome assets.",
+        GET_ASSET_CMD: "Get the path to a local asset.",
+        INSERT_CMD: "Add local asset to the config file."
     }
 
     sps = {}
@@ -88,41 +93,56 @@ def build_argparser():
                 help="URL to use for the genome_server attribute in config file."
                 " Defaults : {}".format(DEFAULT_SERVER))
     sps[BUILD_CMD] = pypiper.add_pypiper_args(
-        sps[BUILD_CMD], groups=None, args=["recover", "config"])
+        sps[BUILD_CMD], groups=None, args=["recover", "config", "new-start"])
 
     # Add any arguments specific to subcommands.
 
-    sps[BUILD_CMD].add_argument(
-        '-i', '--input', dest='input', required=True,
-        help='Local path or URL to genome sequence file in .fa, .fa.gz, '
-             'or .2bit format.')
-    sps[BUILD_CMD].add_argument(
-        '-n', '--name', dest='name', required=False,
-        help='Name of the genome to build. If omitted, refgenie will use the '
-             'basename of the file specified in --input.')
-    sps[BUILD_CMD].add_argument(
-        '-a', '--annotation', dest='annotation', required=False,
-        help='Path to GTF gene annotation file.')
     sps[BUILD_CMD].add_argument(
         "-d", "--docker", action="store_true",
         help="Run all commands in the refgenie docker container.")
+
+    sps[BUILD_CMD].add_argument(
+        '-v', '--volumes', nargs="+", required=False, default=None,
+        help='If using docker, also mount these folders as volumes')
+
     sps[BUILD_CMD].add_argument(
         '-o', '--outfolder', dest='outfolder', required=False, default=None,
-        help='Override the default path to genomes folder, which is to use the '
+        help='Override the default path to genomes folder, which is the '
              'genome_folder attribute in the genome configuration file.')
 
-    for cmd in [PULL_CMD, GET_ASSET_CMD]:
+    for cmd in [PULL_CMD, GET_ASSET_CMD, BUILD_CMD, INSERT_CMD]:
         sps[cmd].add_argument(
             "-g", "--genome", required=True,
             help="Reference assembly ID, e.g. mm10")
         sps[cmd].add_argument(
             "-a", "--asset", required=True, nargs='+',
-            help="Name of asset, a key in a genome config file")
+            help="Name of one or more assets (keys in genome config file)")
 
     sps[PULL_CMD].add_argument(
         "-u", "--no-untar", action="store_true",
         help="Do not extract tarballs.")
 
+    sps[INSERT_CMD].add_argument(
+        "-p", "--path", required=True,
+        help="Relative path to asset")
+
+    # Finally, arguments to the build command to give the files needed to do
+    # the building. These should eventually move to a more flexible system that
+    # doesn't require them to be hard-coded here in order to be recognized
+
+    for arg in BUILD_SPECIFIC_ARGS:
+        sps[BUILD_CMD].add_argument(
+            "--{arg}".format(arg=arg), required=False, help=SUPPRESS)
+
+    # sps[BUILD_CMD].add_argument(
+    #     '--fasta', required=False, help=SUPPRESS)
+    # help='Local path or URL to genome sequence file in .fa, .fa.gz, '
+    #          'or .2bit format.'
+    # sps[BUILD_CMD].add_argument(
+    #        '--gtf', required=False, help=SUPPRESS)
+    # help='Path to GTF gene annotation file.'
+
+
     return parser
 
 
@@ -141,15 +161,15 @@ def copy_or_download_file(input_string, outfolder):
     return result_file, " ".join(parts)
 
 
-def convert_file(input_file, output_file, conversions):
+def convert_file(input_fasta, output_file, conversions):
     """
     Given an input file, output file, and a list of conversions, gives the appropriate output file.
     
     :param str output_file: Path to local output file you want to create
     :param dict conversions: A dictionary of shell commands to convert files of a given type.
     """
-    form = {"INPUT": input_file, "OUTPUT": output_file}
-    _, ext = os.path.splitext(input_file)
+    form = {"INPUT": input_fasta, "OUTPUT": output_file}
+    _, ext = os.path.splitext(input_fasta)
     if ext in conversions:
         return conversions[ext].format(**form)
 
@@ -163,6 +183,27 @@ def default_config_file():
     return os.path.join(os.path.dirname(__file__), "refgenie.yaml")
 
 
+def get_asset_vars(genome, asset_key, outfolder, specific_args=None):
+    """
+    Gives a dict with variables used to populate an asset path.
+    """
+    asset_outfolder = os.path.join(outfolder, asset_key)
+    asset_vars = {"genome": genome,
+                  "asset": asset_key,
+                  "asset_outfolder": asset_outfolder}
+    if specific_args:
+        asset_vars.update(specific_args)
+    return asset_vars
+
+
+def refgenie_add(rgc, args):
+    outfolder = os.path.abspath(os.path.join(rgc.genome_folder, args.genome))
+    asset_vars = get_asset_vars(args.genome, args.asset, outfolder)
+    rgc.update_assets(args.genome, args.asset, {"path": args.path.format(**asset_vars)})
+    # Write the updated refgenie genome configuration
+    rgc.write()
+
+
 def refgenie_build(rgc, args):
     """
     Runs the refgenie build recipe.
@@ -171,28 +212,32 @@ def refgenie_build(rgc, args):
     :param argparse.Namespace args: parsed command-line options/arguments
     """
 
-    if args.name:
-        genome_name = args.name
+    # Build specific args
+    specific_args = {k: getattr(args, k) for k in BUILD_SPECIFIC_ARGS}
+
+    if args.genome:
+        genome = args.genome
     else:
-        genome_name = os.path.basename(args.input)
+        # This can probably be eliminated now that with flexible building
+        genome = os.path.basename(args.input)
         # eliminate extensions to get canonical genome name.
         for strike in [".fasta.gz$", ".fa.gz$", ".fasta$", ".fa$", ".gz$", ".2bit$"]:
-            genome_name = re.sub(strike, "", genome_name)
+            genome = re.sub(strike, "", genome)
 
-    _LOGGER.info("Using genome name: {}".format(genome_name))
+    _LOGGER.info("Using genome name: {}".format(genome))
 
     if not hasattr(args, "outfolder") or not args.outfolder:
         # Default to genome_folder
         _LOGGER.debug("No outfolder provided, using genome config.")
         args.outfolder = rgc.genome_folder
 
-    outfolder = os.path.abspath(os.path.join(args.outfolder, genome_name))
+    outfolder = os.path.abspath(os.path.join(args.outfolder, genome))
     if not _writeable(outfolder):
         _LOGGER.error("Insufficient permissions to write to output folder: {}".
                       format(outfolder))
         return
 
-    _LOGGER.info("Output to: {} {} {}".format(genome_name, args.outfolder, outfolder))
+    _LOGGER.info("Output to: {} {} {}".format(genome, args.outfolder, outfolder))
     _LOGGER.debug("Default config file: {}".format(default_config_file()))
 
     if args.config_file and not os.path.isfile(args.config_file):
@@ -200,161 +245,83 @@ def refgenie_build(rgc, args):
                       format(args.config_file))
         args.config_file = default_config_file()
 
-    pm = pypiper.PipelineManager(name="refgenie", outfolder=outfolder, args=args)
-    tk = pypiper.NGSTk(pm=pm)
-    tools = pm.config.tools  # Convenience alias
-    index = pm.config.index
-    param = pm.config.param
-
-    # pm.make_sure_path_exists(outfolder)
-    conversions = {}
-    conversions[".2bit"] = "twoBitToFa {INPUT} {OUTPUT}"
-    conversions[".gz"] = tk.ziptool + " -cd {INPUT} > {OUTPUT}"
-
-    # Copy fasta file to genome folder structure
-    local_raw_fasta = genome_name + ".fa"
-    raw_fasta = os.path.join(outfolder, local_raw_fasta)
-
-    input_file, cmd = copy_or_download_file(args.input, outfolder)
-    pm.run(cmd, input_file)
-
-    container = None
-    if args.docker:
-        # Set up some docker stuff
-        pm.get_container("nsheff/refgenie", outfolder)
-        # of = os.path.abspath(outfolder)
-        # outfolder = of
-        # cmd = "docker run -itd"
-        # cmd += " -v " + of + ":" + of
-        # cmd += " nsheff/refgenie"
-        # container = pm.checkprint(cmd).rstrip()
-        # _LOGGER.info("Using docker container: " + container)
-        # pm.atexit_register(remove_container, container)
-
-    cmd = convert_file(input_file, raw_fasta, conversions)
-    if cmd:
-        pm.run(cmd, raw_fasta, container=pm.container)
-
-    cmd = tools.samtools + " faidx " + raw_fasta
-    pm.run(cmd, raw_fasta + ".fai", container=pm.container)
-
-    # Determine chromosome sizes
-    fai_file = raw_fasta + ".fai"
-    # symlinks should be relative so folders are portable.
-    loc_chrom_sizes_file = genome_name + ".chrom.sizes"
-    chrom_sizes_file = os.path.join(outfolder, loc_chrom_sizes_file)
-    chrom_sizes_alias = os.path.join(outfolder, genome_name + ".chromSizes")
-    cmd = "cut -f 1,2 " + fai_file + " > " + chrom_sizes_file
-    cmd2 = "ln -s " + loc_chrom_sizes_file + " " + chrom_sizes_alias
-    pm.run([cmd, cmd2], chrom_sizes_alias, container=pm.container)
-
-    # Copy annotation file (if any) to folder structure
-    if args.annotation:
-        annotation_file_unzipped = os.path.join(outfolder, genome_name + ".gtf")
-        annotation_file, cmd = copy_or_download_file(args.annotation, outfolder)
-        pm.run(cmd, annotation_file)
-
-        cmd = convert_file(annotation_file, annotation_file_unzipped, conversions)
-        pm.run(cmd, annotation_file_unzipped)
-
-    #   cmd = "cp " + args.annotation + " " + annotation_file
-    #   cmd2 = tk.ziptool + " -d " + annotation_file
-    #   pm.run([cmd, cmd2], annotation_file_unzipped)
-
-    else:
-        _LOGGER.debug("* No GTF gene annotations provided. Skipping this step.")
-
     def path_data(root, c):
         return {"path": os.path.relpath(root, c.genome_folder)}
 
-    # Bowtie indexes
-    if index.bowtie2:
-        asset_key = "indexed_bowtie2"
-        folder = os.path.join(outfolder, asset_key)
-        tk.make_dir(folder)
-        target = os.path.join(folder, "completed.flag")
-        cmd1 = "ln -sf ../" + local_raw_fasta + " " + folder
-        cmd2 = tools.bowtie2build + " " + raw_fasta + " " + os.path.join(folder, genome_name)
-        cmd3 = "touch " + target
-        pm.run([cmd1, cmd2, cmd3], target, container=pm.container)
+
+    def build_asset(genome, asset_key, asset_build_package, outfolder, specific_args):
+        """
+        Builds assets with pypiper and updates a genome config file.
+
+        This function actually run the build commands in a given build package,
+        and then update the refgenie config file.
+
+        :param str genome: The assembly key; e.g. 'mm10'.
+        :param str asset_key: The unique asset identifier; e.g. 'bowtie2_index'
+        :param dict asset_build_package: A dict (see examples) specifying lists
+            of required inputs, commands to run, and outputs to register as
+            assets.
+        """
+        _LOGGER.debug("Asset build package: " + str(asset_build_package))
+        asset_vars = get_asset_vars(genome, asset_key, outfolder, specific_args)
+        asset_outfolder = os.path.join(outfolder, asset_key)
+
+        _LOGGER.debug(str([x.format(**asset_vars) for x in asset_build_package["command_list"]]))
+
+        tk.make_dir(asset_outfolder)
+        target = os.path.join(asset_outfolder, "build_complete.flag")
+        command_list_populated = [x.format(**asset_vars) for x in asset_build_package["command_list"]]
+
+        touch_target = "touch {target}".format(target=target)
+        command_list_populated.append(touch_target)
+        
+        _LOGGER.debug("Command list populated: " + str(command_list_populated))
+
+        pm.run(command_list_populated, target, container=pm.container)
         # Add index information to rgc
-        rgc.update_genomes(genome_name, asset_key, path_data(folder, rgc))
+        for asset_key, relative_path in asset_build_package["assets"].items():
+            rgc.update_assets(genome, asset_key, {"path": relative_path.format(**asset_vars)})
 
         # Write the updated refgenie genome configuration
         rgc.write()
 
-    # Bismark index - bowtie2
-    if index.bismark_bt2:
-        asset_key = "indexed_bismark_bt2"
-        folder = os.path.join(outfolder, asset_key)
-        tk.make_dir(folder)
-        target = os.path.join(folder, "completed.flag")
-        cmd1 = "ln -sf ../" + local_raw_fasta + " " + folder
-        cmd2 = tools.bismark_genome_preparation + " --bowtie2 " + folder
-        cmd3 = "touch " + target
-        pm.run([cmd1, cmd2, cmd3], target, container=pm.container)
-        rgc.update_genomes(genome_name, asset_key, path_data(folder, rgc))
-        rgc.write()
+    pm = pypiper.PipelineManager(name="refgenie", outfolder=outfolder, args=args)
+    tk = pypiper.NGSTk(pm=pm)
 
-    # Bismark index - bowtie1
-    if index.bismark_bt1:
-        asset_key = "indexed_bismark_bt1"
-        folder = os.path.join(outfolder, asset_key)
-        tk.make_dir(folder)
-        target = os.path.join(folder, "completed.flag")
-        cmd1 = "ln -sf ../" + local_raw_fasta + " " + folder
-        cmd2 = tools.bismark_genome_preparation + " " + folder
-        cmd3 = "touch " + target
-        pm.run([cmd1, cmd2, cmd3], target, container=pm.container)
-        rgc.update_genomes(genome_name, asset_key, path_data(folder, rgc))
-        rgc.write()
+    if args.docker:
+        # Set up some docker stuff
+        if args.volumes:
+            volumes = volumes.append(outfolder)
+        else:
+            volumes = outfolder
 
-    # Epilog meth calling
-    if index.epilog:
-        asset_key = "indexed_epilog"
-        folder = os.path.join(outfolder, asset_key)
-        tk.make_dir(folder)
-        target = os.path.join(folder, "completed.flag")
-        cmd1 = "ln -sf ../" + local_raw_fasta + " " + folder
-        cmd2 = tools.epilog_indexer + " -i " + raw_fasta
-        cmd2 += " -o " + os.path.join(folder, genome_name + "_" + param.epilog.context + ".tsv")
-        cmd2 += " -s " + param.epilog.context  # context
-        cmd2 += " -t"
-        cmd3 = "touch " + target
-        pm.run([cmd1, cmd2, cmd3], target, container=pm.container)
-        rgc.update_genomes(genome_name, asset_key, path_data(folder, rgc))
-        rgc.write()
+    for asset_key in args.asset:
+        if asset_key in asset_build_packages.keys():
+            asset_build_package = asset_build_packages[asset_key]
+            _LOGGER.debug(specific_args)
+            required_inputs = ", ".join(asset_build_package["required_inputs"])
+            _LOGGER.info("Inputs required to build '{}': {}".format(asset_key, required_inputs))
+            for required_input in asset_build_package["required_inputs"]:
+                if not specific_args[required_input]:
+                    raise ValueError("Argument '{}' is required to build asset '{}', but not provided".format(required_input, asset_key))
 
-    if index.hisat2:
-        asset_key = "indexed_hisat2"
-        folder = os.path.join(outfolder, asset_key)
-        tk.make_dir(folder)
-        target = os.path.join(folder, "completed.flag")
-        cmd1 = "ln -sf ../" + local_raw_fasta + " " + folder
-        cmd2 = tools.hisat2build + " " + raw_fasta + " " + os.path.join(folder, genome_name)
-        cmd3 = "touch " + target
-        pm.run([cmd1, cmd2, cmd3], target, container=pm.container)
-        rgc.update_genomes(genome_name, asset_key, path_data(folder, rgc))
-        rgc.write()
-
-    # Kallisto should index transcriptome
-    # So it doesn't make sense to run these at the same time as the others.
-    if index.kallisto:
-        asset_key = "indexed_kallisto"
-        folder = os.path.join(outfolder, asset_key)
-        tk.make_dir(folder)
-        target = os.path.join(folder, "completed.flag")
-        cmd2 = tools.kallisto + " index -i " + os.path.join(folder, genome_name + "_kallisto_index.idx")
-        cmd2 += " " + raw_fasta
-        cmd3 = "touch " + target
-        pm.run([cmd2, cmd3], target, container=pm.container)
-        rgc.update_genomes(genome_name, asset_key, path_data(folder, rgc))
-        rgc.write()
+            for required_asset in asset_build_package["required_assets"]:
+                try:
+                    if not rgc.get_asset(args.genome, required_asset):
+                        raise ValueError("Asset '{}' is required to build asset '{}', but not provided".format(required_asset, asset_key))                    
+                except refgenconf.exceptions.MissingGenomeError:
+                        raise ValueError("Asset '{}' is required to build asset '{}', but not provided".format(required_asset, asset_key))                    
+            if args.docker:
+                pm.get_container(asset_build_package["container"], volumes)
+            build_asset(args.genome, asset_key, asset_build_package, outfolder, specific_args)
+            _LOGGER.info("Finished building asset '{}'".format(asset_key))
+        else:
+            _LOGGER.warn("Recipe does not exist for asset '{}'".format(asset_key))
 
     pm.stop_pipeline()
 
 
-def refgenie_init(genome_config_path, genome_server=DEFAULT_SERVER):
+def refgenie_init(genome_config_path, genome_server=DEFAULT_SERVER, config_version=REQ_CFG_VERSION):
     """
     Initialize a genome config file.
     
@@ -365,6 +332,7 @@ def refgenie_init(genome_config_path, genome_server=DEFAULT_SERVER):
 
     # Set up default 
     rgc = RefGenConf(OrderedDict({
+        CFG_VERSION_KEY: config_version,
         CFG_FOLDER_KEY: os.path.dirname(os.path.abspath(genome_config_path)),
         CFG_SERVER_KEY: genome_server,
         CFG_GENOMES_KEY: None
@@ -383,10 +351,14 @@ def _exec_list(rgc, remote):
     if remote:
         pfx = "Remote"
         assemblies, assets = rgc.list_remote()
+        recipes = None  # Not implemented
     else:
         pfx = "Local"
         assemblies, assets = rgc.list_local()
-    return pfx, assemblies, assets
+        # also get recipes
+        recipes = ", ".join(list(asset_build_packages.keys()))
+
+    return pfx, assemblies, assets, recipes
 
 
 def perm_check_x(file_to_check, message_tag):
@@ -417,7 +389,7 @@ def main():
     global _LOGGER
     _LOGGER = logmuse.logger_via_cli(args)
     logmuse.logger_via_cli(args, name=refgenconf.__name__)
-
+    _LOGGER.info("refgenie {}".format(__version__))
     _LOGGER.debug("Args: {}".format(args))
 
     if not args.command:
@@ -445,7 +417,16 @@ def main():
 
     elif args.command == GET_ASSET_CMD:
         _LOGGER.debug("getting asset: '{}/{}'".format(args.genome, args.asset))
-        return " ".join([rgc.get_asset(args.genome, asset) for asset in args.asset])
+        print(" ".join([rgc.get_asset(args.genome, asset) for asset in args.asset]))
+        return
+
+    elif args.command == INSERT_CMD:
+        if len(args.asset) > 1:
+            raise NotImplementedError("Can only add 1 asset at a time")
+        else:
+            # recast from list to str
+            args.asset = args.asset[0]
+        refgenie_add(rgc, args)
 
     elif args.command == PULL_CMD:
         outdir = rgc[CFG_FOLDER_KEY]
@@ -461,8 +442,10 @@ def main():
         rgc.pull_asset(args.genome, args.asset, gencfg, unpack=not args.no_untar)
 
     elif args.command in [LIST_LOCAL_CMD, LIST_REMOTE_CMD]:
-        pfx, genomes, assets = _exec_list(rgc, args.command == LIST_REMOTE_CMD)
+        pfx, genomes, assets, recipes = _exec_list(rgc, args.command == LIST_REMOTE_CMD)
         _LOGGER.info("{} genomes: {}".format(pfx, genomes))
+        if args.command != LIST_REMOTE_CMD:  # Not implemented yet
+            _LOGGER.info("{} recipes: {}".format(pfx, recipes))
         _LOGGER.info("{} assets:\n{}".format(pfx, assets))
 
 
